@@ -3,17 +3,24 @@ import logging
 import venusian
 from pyramid.exceptions import ConfigurationError
 from pyramid.httpexceptions import HTTPNotFound
+from pyramid.renderers import null_renderer
 from pyramid.response import Response
 from pyramid.security import NO_PERMISSION_REQUIRED
 
-from pyramid_rpc.api import MapplyViewMapper
-from pyramid_rpc.api import ViewMapperArgsInvalid
+from pyramid_rpc.compat import is_nonstr_iter
+from pyramid_rpc.compat import text_type
 from pyramid_rpc.compat import xmlrpclib
+from pyramid_rpc.mapper import MapplyViewMapper
+from pyramid_rpc.mapper import ViewMapperArgsInvalid
+from pyramid_rpc.util import combine
 
 
 log = logging.getLogger(__name__)
 
 
+_marker = object()
+
+
 class XmlRpcError(xmlrpclib.Fault):
     faultCode = None
     faultString = None
@@ -62,20 +69,69 @@ def exception_view(exc, request):
     return response
 
 
-def xmlrpc_renderer(info):
-    def _render(value, system):
-        request = system.get('request')
-        if request is not None:
-            response = request.response
-            ct = response.content_type
-            if ct == response.default_content_type:
-                response.content_type = 'text/xml'
+def make_response(request, result):
+    response = request.response
 
-            return xmlrpclib.dumps((value,), methodresponse=True)
-    return _render
+    ct = response.content_type
+    if ct == response.default_content_type:
+        response.content_type = 'text/xml'
+
+    response.body = (
+        xmlrpclib.dumps(
+            (result,), methodresponse=True
+        ).encode(response.charset)
+    )
+    return response
 
 
-def setup_xmlrpc(request):
+class xmlrpc_view(object):
+    """ Decorator that wraps a view and converts the result into a valid
+    JSON-RPC Response object.
+
+    """
+    def __init__(self, wrapped):
+        self.wrapped = wrapped
+
+    def __call__(self, context, request):
+        result = self.wrapped(context, request)
+        if not request.is_response(result):
+            result = make_response(request, result)
+        return result
+
+
+class EndpointPredicate(object):
+    def __call__(self, info, request):
+        # find the endpoint info
+        key = info['route'].name
+        endpoint = request.registry.rpc_endpoints[key]
+
+        # parse the request body
+        setup_request(endpoint, request)
+
+        # update request with endpoint information
+        request.rpc_endpoint = endpoint
+
+        # Always return True so that even if it isn't a valid RPC it
+        # will fall through to the notfound_view which will still
+        # return a valid XML-RPC response.
+        return True
+
+
+class MethodPredicate(object):
+    def __init__(self, method):
+        self.method = method
+
+    def __call__(self, context, request):
+        return getattr(request, 'rpc_method') == self.method
+
+
+class Endpoint(object):
+    def __init__(self, name, default_mapper):
+        self.name = name
+        self.default_mapper = default_mapper
+
+
+def setup_request(endpoint, request):
     try:
         params, method = xmlrpclib.loads(request.body)
     except Exception:
@@ -88,40 +144,47 @@ def setup_xmlrpc(request):
         raise XmlRpcMethodNotFound
 
 
-def add_xmlrpc_endpoint(self, name, *args, **kw):
+def add_xmlrpc_endpoint(config, name, *args, **kw):
     """Add an endpoint for handling XML-RPC.
 
-    name
+    ``name``
 
         The name of the endpoint.
 
+    ``default_mapper``
+
+        A default view mapper that will be passed as the ``mapper``
+        argument to each of the endpoint's methods.
+
     A XML-RPC method also accepts all of the arguments supplied to
     Pyramid's ``add_route`` method.
 
     """
-    def xmlrpc_endpoint_predicate(info, request):
-        # potentially setup either rpc v1 or v2 from the parsed body
-        setup_xmlrpc(request)
+    default_mapper = kw.pop('default_mapper', MapplyViewMapper)
+
+    endpoint = Endpoint(
+        name,
+        default_mapper=default_mapper,
+    )
+
+    config.registry.rpc_endpoints[name] = endpoint
 
-        # Always return True so that even if it isn't a valid RPC it
-        # will fall through to the notfound_view which will still
-        # return a valid XML-RPC response.
-        return True
     predicates = kw.setdefault('custom_predicates', [])
-    predicates.append(xmlrpc_endpoint_predicate)
-    self.add_route(name, *args, **kw)
-    self.add_view(exception_view, route_name=name, context=Exception,
+    predicates.append(EndpointPredicate())
+
+    config.add_route(name, *args, **kw)
+    config.add_view(exception_view, route_name=name, context=Exception,
                   permission=NO_PERMISSION_REQUIRED)
 
 
-def add_xmlrpc_method(self, view, **kw):
+def add_xmlrpc_method(config, view, **kw):
     """Add a method to a XML-RPC endpoint.
 
-    endpoint
+    ``endpoint``
 
         The name of the endpoint.
 
-    method
+    ``method``
 
         The name of the method.
 
@@ -133,26 +196,51 @@ def add_xmlrpc_method(self, view, **kw):
     behavior simply set the ``mapper`` argument to None or another
     view mapper.
 
+    .. note::
+
+       An endpoint **must** be defined before methods may be added.
+
     """
-    endpoint = kw.pop('endpoint', kw.pop('route_name', None))
-    if endpoint is None:
+    endpoint_name = kw.pop('endpoint', kw.pop('route_name', None))
+    if endpoint_name is None:
         raise ConfigurationError(
             'Cannot register a XML-RPC endpoint without specifying the '
             'name of the endpoint.')
 
+    endpoint = config.registry.rpc_endpoints.get(endpoint_name)
+    if endpoint is None:
+        raise ConfigurationError(
+            'Could not find an endpoint with the name "%s".' % endpoint_name)
+
     method = kw.pop('method', None)
     if method is None:
         raise ConfigurationError(
             'Cannot register a XML-RPC method without specifying the '
             '"method"')
 
-    def xmlrpc_method_predicate(context, request):
-        return getattr(request, 'rpc_method', None) == method
+    mapper = kw.pop('mapper', _marker)
+    if mapper is _marker:
+        # only override mapper if not supplied
+        mapper = endpoint.default_mapper
+    kw['mapper'] = mapper
+
+    decorator = kw.get('decorator', None)
+    if decorator is None:
+        decorator = xmlrpc_view
+    else:
+        if not is_nonstr_iter(decorator):
+            decorator = (decorator,)
+        # we want to apply the view_wrapper first, then the other decorators
+        # and combine() reverses the order, so ours goes last
+        decorators = list(decorator) + [xmlrpc_view]
+        decorator = combine(*decorators)
+    kw['decorator'] = decorator
+
     predicates = kw.setdefault('custom_predicates', [])
-    predicates.append(xmlrpc_method_predicate)
-    kw.setdefault('mapper', MapplyViewMapper)
-    kw.setdefault('renderer', 'pyramid_rpc:xmlrpc')
-    self.add_view(view, route_name=endpoint, **kw)
+    predicates.append(MethodPredicate(method))
+
+    kw['renderer'] = null_renderer
+    config.add_view(view, route_name=endpoint_name, **kw)
 
 
 class xmlrpc_method(object):
@@ -204,8 +292,10 @@ def includeme(config):
     - ``add_xmlrpc_method``: Add a method to a XML-RPC endpoint.
 
     """
+    if not hasattr(config.registry, 'rpc_endpoints'):
+        config.registry.rpc_endpoints = {}
+
     config.add_directive('add_xmlrpc_endpoint', add_xmlrpc_endpoint)
     config.add_directive('add_xmlrpc_method', add_xmlrpc_method)
-    config.add_renderer('pyramid_rpc:xmlrpc', xmlrpc_renderer)
     config.add_view(exception_view, context=xmlrpclib.Fault,
                     permission=NO_PERMISSION_REQUIRED)
